import re, sys, os
from BeautifulSoup import BeautifulSoup
from JSFilter import JSFilter
from thomas import Bayes, WebPageTokenizer

from constants import *

_write_pattern = re.compile("document.write(.*)")

_bayesfilter = Bayes(WebPageTokenizer())
_bayesfilter.load(os.path.join("..", "data", "bayesdata.dat"))

def _should_parse_js():
    return len(sys.argv) == 2 and sys.argv[1] == "js"

class HTTPReply:
    """We parse the HTTP response and put in that class"""
    def __init__(self, httpresponse, userid):
        self.uid = userid
        self.status = httpresponse.status
        self.reason = httpresponse.reason
        self.httpresponse = httpresponse
        self.getheaders()
    
    def getheaders(self):
        self.headers = dict((hdr) for hdr in self.httpresponse.getheaders())
        if self.headers.has_key("transfer-encoding"):
            del self.headers["transfer-encoding"]
        self.content_header = self.get_header("content-type")
    
    def feed_body(self, bytes):
        self.body = self.httpresponse.read(bytes)
    
    def feed_all(self):
        self.js_outputs = []
        if self._is_content_html():
            tmp = ""
            if self.body:
                tmp = self.body    
            self.body = BeautifulSoup()
            self.body.feed(tmp + self.httpresponse.read())
            js_filter = JSFilter()
            scripts = self.body.fetch("script")
            if _should_parse_js():
                for script in scripts:
                    if len(script.contents) == 0:
                        continue
                    script_data = script.contents[0]
                    js_filter.execute_script(script_data, 1.1)
                    script_data = re.sub(_write_pattern, "", script_data)
                    self.js_outputs.append(js_filter.data)
                    script.contents[0] = script_data
            
            #determine categories the page belongs to            
            bayes_guess = _bayesfilter.guess(self.body)
            #print "Bayes Guess", bayes_guess
            probs = [0.7, 0.6, 0.5, 0.4]
            
            if len(bayes_guess) != 1:
                bayes_guess = [(category, prob) for category, prob in bayes_guess
                                                    if prob != 0.9999]
            
            for p in probs:
                tmp_bayes = [(category,prob) 
                                    for category, prob in bayes_guess 
                                        if prob > p]
                if len(tmp_bayes) > 0:
                    bayes_guess = tmp_bayes
                    break
            
            for cat, prob in bayes_guess:
                if (cat == "porn" or cat == "gambling") and prob < 0.6:
                    bayes_guess.remove((cat, prob))
            
            self.categories = [category for category, prob in bayes_guess]
            #print self.categories
            
        else:
            self.body += self.httpresponse.read()
        
    
    def serialize(self):
        """It serializes HTTP Response so that it can be sent as a string to the browser"""
        seristring = "HTTP/1.1 %s %s \r\n" % (self.status, self.reason)
        for i in self.headers.items():
            seristring += "%s: %s\r\n" % (i[0].title(), i[1])
        if _should_parse_js() and self._is_content_html():
            scripts = self.body.fetch("script")
            for ii in xrange(len(scripts)):
                try: scripts[ii].contents[0] += ";document.write('%s');" % self.js_outputs[ii]
                except: continue
           
        seristring += "\r\n" + str(self.body)
        return seristring
    
    def get_header(self, header):
        lcheader = header.lower()
        if self.headers.has_key(lcheader):
            return self.headers[lcheader]
        else:
            return None
    
        
    def replace_header(self, headerName, headerValue):
        self.headers[headerName.lower()] = headerValue
        
    def delete_header(self, headerName, headerValue):
        lcheader = headerName.lower()
        if self.headers.has_key(lcheader):
            if headerValue == "" or headerValue == self.headers[lcheader]:
                del self.headers[lcheader]
                
    def _is_content_html(self):
        return self.content_header and \
               re.match("text\/(html|xhtml)", self.content_header)
