""" This is the plugin engine responsible for plugins"""

import imp, os, re, cPickle, time

from sqlobject import SQLObjectNotFound

from constants import *
from config import *
from model import ICPlugin, ICRule
import plugins

plugMap = {}
ruleMap = {}
ruleDBMap = {}

#TODO: Add error checking to the whole module!!

def load_plugins():
    """It reads ICPlugin in our database and then initiliazes respective modules"""
    for i in ICPlugin.select():
        add_plugin(i.pluginName)
    
def load_rules():
    """ It reads all ICRule in our database, then creates rules from the parameters"""
    for i in ICRule.select():
        add_or_update_rule(i.id, True)

def check_args(pluginName, args):
    """ Called by the web interface to provide feedback to the user """
    try:
        plugin_class = plugMap[pluginName]
        plugin_class(args)  ## may throw an exception 
        return True, ""
    except ParamError, args:
        return False, args
    
def add_or_update_rule(ruleId, start_up=False):
    if not start_up:
        time.sleep(2)
    rule = ICRule.get(ruleId)
    rule.sync()
    print rule.ruleName, "updated", "new urls", rule.appliedURLs
    plugin_class = plugMap[rule.plugin.pluginName]
    ruleMap[ruleId] = plugin_class(rule.parameters)
    ruleDBMap[ruleId] = rule
    
def delete_rule(ruleId):
    del ruleMap[ruleId]
    del ruleDBMap[ruleId]

def add_plugin(pluginName):
    fp, pathname, description = imp.find_module(pluginName, plugins.__path__)
    try:
        dmod = imp.load_module(pluginName, fp, pathname, description)
        plugMap[pluginName] = getattr(dmod, pluginName)
    finally:
        # Since we may exit via an exception, close fp explicitly.
        if fp:
            fp.close()
    plugin_class = plugMap[pluginName]
    try:
        plugin = ICPlugin.byPluginName(pluginName)
        plugin.arguments = plugin_class.Args
        plugin.description = plugin_class.Desc
        plugin.applyTime = plugin_class.ApplyTime
    except SQLObjectNotFound:
        plugin = ICPlugin(pluginName=pluginName, 
                          arguments=plugin_class.Args, 
                          description=plugin_class.Desc,
                          applyTime=plugin_class.ApplyTime)
    
def get_args_plugin(pluginName):
    return plugMap[pluginName].Args
