""" This is the core part of the proxy

This core gets the requests then parses the request with the rules then then sends the requests,then gets the response and parses it,at last it sends the response to web browser
For all requests it creates a new thread then associates this thread with a connection
and also responsible for logging and executing the rules
"""

import SocketServer, httplib, re, time, thread, datetime
from sqlobject import SQLObjectNotFound

from constants import *
from config import *
from HTTPRequest import *
from HTTPReply import *
from ProxyListener import *
from model import *
import PlugCore
import Auth

def redirect(location):
    headers = "HTTP/1.1 302 Found\r\nLocation: @LOCATION@\r\nContent-Type: text/html\r\nContent-Length: @LEN@\r\n\r\n"
    body = """<HTML><HEAD><meta http-equiv="content-type" content="text/html;charset=utf-8"><TITLE>302 Moved</TITLE></HEAD><BODY><H1>302 Moved</H1>You are being redirected to <A HREF="@LOCATION@">here</A>.</BODY></HTML>\r\n"""
    body = body.replace("@LOCATION@", location)
    headers = headers.replace("@LOCATION@", location)
    headers = headers.replace("@LEN@", str(len(body)))
    
    return headers + body
    
_country_type = ['ac', 'ad', 'ae', 'af', 'ag', 'ai', 'al', 'am', 'an', 'ao', 'aq',
                 'ar', 'as', 'at', 'au', 'aw', 'az', 'ba', 'bb', 'bd', 'be', 'bf',
                 'bg', 'bh', 'bi', 'bj', 'bm', 'bn', 'bo', 'br', 'bs', 'bt', 'bv',
                 'bw', 'by', 'bz', 'ca', 'cc', 'cd', 'cf', 'cg', 'ch', 'ci', 'ck',
                 'cl', 'cm', 'cn', 'co', 'cr', 'cu', 'cv', 'cx', 'cy', 'cz', 'de',
                 'dj', 'dk', 'dm', 'do', 'dz', 'ec', 'ee', 'eg', 'er', 'es', 'et',
                 'eu', 'fi', 'fj', 'fk', 'fm', 'fo', 'fr', 'ga', 'gd', 'ge', 'gf',
                 'gg', 'gh', 'gi', 'gl', 'gm', 'gn', 'gp', 'gq', 'gr', 'gs', 'gt',
                 'gu', 'gw', 'gy', 'hk', 'hm', 'hn', 'hr', 'ht', 'hu', 'id', 'ie',
                 'il', 'im', 'in', 'io', 'iq', 'ir', 'is', 'it', 'je', 'jm', 'jo',
                 'jp', 'ke', 'kg', 'kh', 'ki', 'km', 'kn', 'kr', 'kw', 'ky', 'kz',
                 'la', 'lb', 'lc', 'li', 'lk', 'lr', 'ls', 'lt', 'lu', 'lv', 'ly',
                 'ma', 'mc', 'md', 'mg', 'mh', 'mk', 'ml', 'mm', 'mn', 'mo', 'mp',
                 'mq', 'mr', 'ms', 'mt', 'mu', 'mv', 'mw', 'mx', 'my', 'mz', 'na',
                 'nc', 'ne', 'nf', 'ng', 'ni', 'nl', 'no', 'np', 'nr', 'nu', 'nz',
                 'om', 'pa', 'pe', 'pf', 'pg', 'ph', 'pk', 'pl', 'pm', 'pn', 'pr',
                 'ps', 'pt', 'pw', 'py', 'qa', 're', 'ro', 'ru', 'rw', 'sa', 'sb',
                 'sc', 'sd', 'se', 'sg', 'sh', 'si', 'sk', 'sl', 'sm', 'sn', 'sr',
                 'st', 'su', 'sv', 'sy', 'sz', 'tc', 'td', 'tf', 'tg', 'th', 'tj',
                 'tk', 'tl', 'tm', 'tn', 'to', 'tr', 'tt', 'tv', 'tw', 'tz', 'ua',
                 'ug', 'uk', 'us', 'uy', 'uz', 'va', 'vc', 've', 'vg', 'vi', 'vn',
                 'vu', 'wf', 'ws', 'ye', 'yt', 'yu', 'za', 'zm', 'zw']
                 
_host_type = ['com', 'net', 'org', 'edu', 'gov', 'biz', 'info', 'name', 'mil', 'pro', 'int']

_apply_times = [ BEFORE_BODY, AFTER_4KB, AFTER_BODY ]

template="<html><body><b>%s</b> blocked access to site %s.<br>%s</body></html>"
 
class ProxyHandler (SocketServer.StreamRequestHandler):
    """It runs for only one connection there is threading"""
    def handle(self):
        ip, port = self.client_address
        self.uid = Auth.check_if_auth(ip)
        request_string = ""
        while True:
            new_s = self.rfile.readline()
            if len(new_s.strip()) == 0:
                break
            request_string += new_s
        request_list = (request_string.splitlines())
        self.httprequest = HTTPRequest(request_list, self.uid)
        if self.uid == None:
            self.wfile.write(redirect("http://" + UI_HOST + ":" + str(UI_PORT) + "/login_user"))
            return;
        if self.httprequest.line.method == "POST":
            self.httprequest.body = self.rfile.read(int(self.httprequest.headers["Content-Length"]))
        self.process()
    
    def finish(self):
        """ For closing connection"""
        self.connection.close()
        if hasattr(self, "httpconn"):
            self.httpconn.close()
    
    def _connect_to(self, netloc):
        """It opens a connection to the netloc parameter and returns the relevant HTTP connection object"""
        i = netloc.find(':')
        if i >= 0:
            return httplib.HTTPConnection(netloc[:i], int(netloc[i+1:]))
        
        return httplib.HTTPConnection(netloc, 80)
        
    def process(self): #TODO: maybe split this into multiple functions
        """ It parses request and response. Also sends and gets them
            It does most of the work in core proxy"""
        fpath  = self.httprequest.line.full_path()
        netloc = self.httprequest.line.netloc
        scheme = self.httprequest.line.scheme
        
        self.request_url = scheme + "://" + netloc + fpath
                    
        if(self.httprequest.headers.has_key("Accept-Encoding")):
            del self.httprequest.headers["Accept-Encoding"]
        if(self.httprequest.headers.has_key("Keep-Alive")):
            del self.httprequest.headers["Keep-Alive"];
        self.httprequest.headers["Proxy-Connection"] = "Close"
        
        for i in PlugCore.ruleMap.iteritems():
            #rule = PlugCore.ruleDBMap[i[0]]
            rule = ICRule.get(i[0])
            if self.rule_applicable(rule, self.httprequest):
                res, msg = i[1].act_on_req(self.httprequest)
                if res == ACTION_BLOCK:
                    self.wfile.write(template %(rule.ruleName, netloc, msg))
                    return

        openTime = datetime.datetime.today().replace(microsecond=0)
        try:
            self.httpconn = self._connect_to(netloc)
            self.httpconn.request(self.httprequest.line.method, fpath, self.httprequest.body, self.httprequest.headers)
        except:
            error_page = "<html><body>Error occured while accessing <b>" + netloc + "</b><br/></body></html>"
            print "External error trying to access", netloc, fpath
            self.wfile.write(error_page)
            return
        
        httpresponse = self.httpconn.getresponse()
        
        self.httpreply = HTTPReply(httpresponse, self.uid)
        
        # Apply rules after header-before body, 
        _flag = False
        for applytime in _apply_times:    
            for i in PlugCore.ruleMap.iteritems():
                rule = ICRule.get(i[0])
                if rule.plugin.applyTime == applytime and self.rule_applicable(rule, self.httpreply):
                    res, msg = i[1].act_on_res(self.httpreply)
                    if res == ACTION_ANSWER:
                        _flag = True
                        break    
                    if res == ACTION_BLOCK:
                        self.wfile.write(template %(rule.ruleName, netloc, msg))
                        return
            if _flag:
                break
            if applytime == BEFORE_BODY:
                self.httpreply.feed_body(4096)
            elif applytime == AFTER_4KB:
                self.httpreply.feed_all()

        httpreply_data = self.httpreply.serialize()
        httprequest_data = self.httprequest.serialize() #Wasteful to use this only for bandwith calculation.

        #try:
        #    domain = ""
        #    for i in _host_type:
        #        if i in netloc:
        #            dn = netloc.split(".")
        #            for j in dn[dn.index(i) - 1:]:
        #                domain += j + "."
        #            break

        #    if domain == "":
        #        for i in _country_type:
        #            if i in netloc:
        #                dn = netloc.split(".")
        #                for j in dn[dn.index(i) - 1:]:
        #                    domain += j + "."
        #                break   

        #    domain = domain[:-1]
            
        #except: domain = netloc
        
        inBand = len(httpreply_data)
        outBand = len(httprequest_data)
        ICUserLog(user=self.uid,
                  location=scheme + "://" + netloc + fpath, 
                  inBandwidth= inBand, outBandwidth= outBand,
                  requestTime=openTime, 
                  urlType=self.httpreply.get_header("content-type"))
        
        try:
            user = ICUser.get(self.uid)
            bandw = ICBandwidth.select(
                                AND(ICBandwidth.q.userID == self.uid, 
                                    ICBandwidth.q.date == openTime.replace(microsecond=0,second=0,minute=0,hour=0)))[0]
            #totalBandwidth in kilobytes
            bandw.totalBandwidth += (inBand + outBand)/1024
        except IndexError:
            ICBandwidth(user = user, date = openTime.replace(microsecond=0,second=0,minute=0,hour=0),
                        totalBandwidth = (inBand + outBand)/1024) 
        
        #print "Connection to", netloc, fpath, " ok, closing"
        self.wfile.write(httpreply_data)

    def rule_applicable(self, rule, cReqRes):
        userOK = False
        groupOK = False
        tagOK = False
        urlOK = False
        
        user = ICUser.get(self.uid)
        
        #Checking if the rule should work for user
        if rule.appliedUsers != []:
            for i in rule.appliedUsers:
                if user.id == i.id:
                    userOK = True
                    break
        else:
            userOK = True

        #Checking if the rule should work for group
        if rule.appliedGroups != []:
            for i in rule.appliedGroups:
                if i in user.groups:
                    groupOK = True
                    break
        else:
            groupOK = True
            
        #Checking if the rule should work for tag
        #print "TAGS", rule.appliedTags
        if rule.appliedTags != []:
            if not hasattr(cReqRes, "categories"):
                tagOK = False
            else:
                #print "valla buradayim", rule.appliedTags, cReqRes.categories
                for i in rule.appliedTags:
                    if i.tagName in cReqRes.categories:
                        tagOK = True
                        break
        else:
            tagOK = True
                
        #Checking if the rule should work for url
        if rule.appliedURLs != []:
            for i in rule.appliedURLs:
                #print i, self.request_url
                if re.search(i, self.request_url):
                    urlOK = True
                    break
        else:
            urlOK = True
        
        #Checking if the rule should not apply for url
        if rule.notappliedURLs != []:
            for i in rule.notappliedURLs:
                if re.search(i, self.request_url):
                    urlOK = False
                    break
                            
        return userOK and groupOK and tagOK and urlOK  
                   
class ThreadedProxyServer (SocketServer.ThreadingMixIn,
                           SocketServer.TCPServer): 
    """It associates a thread with a ProxyHandler instance"""
    def init_rules(self):
        """It initializes the plugins and rules  """
        PlugCore.load_plugins()
        PlugCore.load_rules()

    def init_listener(self):
        self.listener = ProxyListener()
        thread.start_new_thread(asyncore.loop, ())
